"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const node_fetch_1 = require("node-fetch");
const qs = require("querystring");
class Provider {
    constructor() {
        this.query = {};
        this.encoding = 'application/json';
    }
    serialiseScopes(req) {
        if (!this.scopes || this.scopes.length === 0) {
            return undefined;
        }
        return this.getScopes(req).join(' ');
    }
    getScopes(req) {
        if (typeof this.scopes === 'function') {
            return this.scopes(this.name, req);
        }
        return this.scopes;
    }
    compileAuthUrl(req, options, redirectUri) {
        const query = Object.assign({ response_type: 'code', redirect_uri: redirectUri, client_id: this.clientId }, this.query);
        const scopes = this.serialiseScopes(req);
        if (scopes) {
            query['scope'] = scopes;
        }
        return options.handler
            .preAuthUrl(query, this, req)
            .then(() => `${this.authUrl}?${qs.stringify(query)}`);
    }
    requestToken(code, redirect_uri) {
        const payload = {
            code,
            redirect_uri,
            client_id: this.clientId,
            client_secret: this.clientSecret,
            grant_type: 'authorization_code',
        };
        return this.sendTokenRequest(payload);
    }
    refreshToken(refresh_token) {
        const payload = {
            refresh_token,
            client_id: this.clientId,
            client_secret: this.clientSecret,
            grant_type: 'refresh_token',
        };
        return this.sendTokenRequest(payload);
    }
    sendTokenRequest(payload) {
        return __awaiter(this, void 0, void 0, function* () {
            let body;
            switch (this.encoding) {
                case 'application/json':
                    body = JSON.stringify(payload);
                    break;
                case 'application/x-www-form-urlencoded':
                    body = qs.stringify(payload);
                    break;
                default:
                    throw new Error(`Unknown encoding type: ${this.encoding}`);
            }
            return node_fetch_1.default(this.tokenUrl, {
                method: 'POST',
                headers: {
                    'Content-Type': this.encoding,
                },
                body,
            }).then((res) => __awaiter(this, void 0, void 0, function* () {
                if (res.status < 200 || res.status >= 300) {
                    throw new Error(`unexpected response code ${res.status}`);
                }
                return res.json();
            }));
        });
    }
    handleCode(h, options, redirectUri) {
        return __awaiter(this, void 0, void 0, function* () {
            const error = this.getCodeError(h.request);
            if (error) {
                return options.handler.onError({ provider: this, error }, h);
            }
            const code = this.extractCode(h.request);
            if (!code) {
                return options.handler.onError({
                    provider: this,
                    error: new CodeError(CodeError.Kind.Missing),
                }, h);
            }
            return this.requestToken(code, redirectUri).then(data => options.handler.onLink({ provider: this, data }, h), error => options.handler.onError({ provider: this, error }, h));
        });
    }
    extractCode(req) {
        return req.query['code'];
    }
    getCodeError(req) {
        const err = req.query['error'];
        if (err == null) {
            return null;
        }
        switch (err) {
            case 'access_denied':
                return new CodeError(CodeError.Kind.Denied);
            default:
                return new CodeError(CodeError.Kind.Unknown);
        }
    }
    /*abstract*/ getProfile(tokens) {
        throw new Error('Not implemented');
    }
}
exports.Provider = Provider;
function registerProvider(server, options, provider) {
    const redirectUri = `${options.baseUrl}/oauth/${provider.name}`;
    server.route({
        method: 'GET',
        path: `/oauth/${provider.name}/request`,
        options: options.requestConfig,
        handler: function (request) {
            return provider
                .compileAuthUrl(request, options, redirectUri)
                .then(url => ({ url }));
        },
    });
    server.route({
        method: 'GET',
        path: `/oauth/${provider.name}`,
        options: options.linkConfig,
        handler: function (request, h) {
            return provider.handleCode(h, options, redirectUri);
        },
    });
}
exports.registerProvider = registerProvider;
class CodeError extends Error {
    constructor(kind) {
        super(kind);
        this.kind = kind;
        Object.setPrototypeOf(this, new.target.prototype);
    }
}
exports.CodeError = CodeError;
(function (CodeError) {
    let Kind;
    (function (Kind) {
        Kind["Denied"] = "Request denied by user";
        Kind["Missing"] = "Code not found in request";
        Kind["Unknown"] = "Unknown error returned from provider";
    })(Kind = CodeError.Kind || (CodeError.Kind = {}));
})(CodeError = exports.CodeError || (exports.CodeError = {}));
